from abc import abstractmethod
from typing import Protocol


class Comparable(Protocol):
    @abstractmethod
    def __lt__(self, other) -> bool: ...
    @abstractmethod
    def __le__(self, other) -> bool: ...
    @abstractmethod
    def __eq__(self, other) -> bool: ...
    @abstractmethod
    def __ne__(self, other) -> bool: ...
    @abstractmethod
    def __gt__(self, other) -> bool: ...
    @abstractmethod
    def __ge__(self, other) -> bool: ...


class SimpleSequence[T](Protocol):
    @abstractmethod
    def __len__(self) -> int: ...
    @abstractmethod
    def __getitem__(self, i: int) -> T: ...
    @abstractmethod
    def __setitem__(self, i: int, value: T): ...
