from __future__ import annotations

from typing import BinaryIO, TextIO

from ..data.fasta import DNARecord

class FastaReader:
    def __init__(self, reader: TextIO | BinaryIO): ...
    @classmethod
    def from_file(cls, path: str) -> FastaReader: ...
    def __iter__(self) -> FastaReader: ...
    def __next__(self) -> DNARecord: ...
